from ..Log import *
from ..Version import __version__, __versionData__
from tempfile import gettempdir
from platform import system
import os
import sys
import urllib.request
import json
import zipfile

__all__ = ["main", "install"]

def install():
    print("Checking compatibility")

    _system = system().lower()

    with urllib.request.urlopen(f"https://deps.libmge.org/?version={__version__}&release={__versionData__['phase']}&os={_system}") as response:
        data = response.read().decode('utf-8')

    data = json.loads(data)

    if 'error' in data:
        LogCritical(data["error"]["title"], data["error"]["message"])
    else:
        print(f"Downloading dependencies from {data['download']}")

        temp_zip_path = os.path.join(gettempdir(), 'deps.zip')

        urllib.request.urlretrieve(data['download'], temp_zip_path)

        print(f"Extracting files")

        with zipfile.ZipFile(temp_zip_path, 'r') as zip_ref:
            zip_ref.extractall(f"{os.path.abspath(os.path.dirname(__file__))}/../")

        Log("Successfully installed")

def main():
    if len(sys.argv) < 2:
        return

    command = sys.argv[1].lower()

    if command == "version":
        print(f"{'MGE' if __versionData__['phase'] == 'Stable' else __versionData__['phase'] + '-MGE'} {__version__}")

    elif command == "deps" and len(sys.argv) >= 3:
        subcommand = sys.argv[2].lower()

        if subcommand == "install":
            install()
        elif subcommand == "versions":
            from ..Platform import Platform
            LogInfo(f"SDL - {Platform.SDL.SDL_version}")
            LogInfo(f"SDLGFX - {Platform.SDL.SDLGFX_version}")
            LogInfo(f"SDLIMAGE - {Platform.SDL.SDLIMAGE_version}")
            LogInfo(f"SDLTTF - {Platform.SDL.SDLTTF_version}")
            LogInfo(f"SDLMIXER - {Platform.SDL.SDLMIXER_version}")



import array
import os
from .Log import LogCritical
from ctypes import cast, POINTER as _P
from ._sdl import sdl2, sdlimage, sdlttf, sdlmixer
from .Time import Time as _Time, fps_to_time
from .Monitors import _update_monitors_datas
from .Constants import All, Pivot2D
from .Mesh import edges, calculate_square_vertices, line_intersection

__all__ = ["_temp", "init", "update", "SetLogicClock", "GetLogicClock", "OpenUrl",
           "AllEvents", "QuitEvent", "WindowEvents", "AutoCalcs2D",
           "_calculate_size", "_calculate_location", "_calculate_object2d", "_calculate_line"]

class _temp:
    LogicClock = 1024

    mouse_motion_tick_time = {"x": _Time(fps_to_time(60)), "y": _Time(fps_to_time(60))}

    Time = {"LogicTime": _Time(fps_to_time(LogicClock))}

    Events = []

    KeyboardState = sdl2.SDL_GetKeyboardState(None)
    KeyboardCache = []
    for num in range(530):
        KeyboardCache.append(True)

    MouseState = [False, False, False, False, False]
    Mouse = {"location": None, "button_cache": [False, False, False, False, False]}

    MouseCursor = 0

    MouseCursors = []

    Button = {"button_active": False}

def init(video=True, audio=False, events=True, controller=False, sensor=False):
    if sdl2.SDL_Init(0) != 0:
        LogCritical("initializing SDL2")

    def sdl2_init(system):
        if sdl2.SDL_InitSubSystem(system) != 0:
            LogCritical(f"initializing the {system} subsystem")

    sdl2_init(sdl2.SDL_INIT_TIMER)
    if events:
        sdl2_init(sdl2.SDL_INIT_EVENTS)
    if video:
        sdl2_init(sdl2.SDL_INIT_VIDEO)
        for num in range(12):
            _temp.MouseCursors.append(sdl2.SDL_CreateSystemCursor(num).contents)
        _update_monitors_datas()
        sdlttf.TTF_Init()
        sdlimage.IMG_Init()
        os.environ["SDL_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR"] = "0"
    if audio:
        sdlmixer.Mix_OpenAudio(360000, sdl2.AUDIO_S16SYS, 2, 1024)
        sdlmixer.Mix_Init()
    if controller:
        sdl2_init(sdl2.SDL_INIT_JOYSTICK)
        sdl2_init(sdl2.SDL_INIT_GAMECONTROLLER)
    # if haptic:
    #    sdl2_init(sdl2.SDL_INIT_HAPTIC)
    if sensor:
        sdl2_init(sdl2.SDL_INIT_SENSOR)

def update():
    _temp.Time["LogicTime"].tickSleep()

    sdl2.SDL_PumpEvents()
    _events = (sdl2.SDL_Event * 10)()
    _num = sdl2.SDL_PeepEvents(cast(_events, _P(sdl2.SDL_Event)), 10, sdl2.SDL_GETEVENT, sdl2.SDL_FIRSTEVENT, sdl2.SDL_LASTEVENT)
    _temp.Events = list(_events)[:_num]

    _keyboard = sdl2.SDL_GetKeyboardState(None)
    _temp.KeyboardState = [bool(_keyboard[_num]) for _num in range(530)]

    _mouse = sdl2.SDL_GetMouseState(None, None)
    _temp.MouseState = [bool(_mouse & (1 << _num)) for _num in range(5)]

    if _temp.Button["button_active"]:
        _temp.Button["button_active"] = False
    sdl2.SDL_SetCursor(_temp.MouseCursors[_temp.MouseCursor])
    _temp.MouseCursor = 0

def SetLogicClock(logic_clock: int):
    _temp.Time["LogicTime"].delta_time = fps_to_time(logic_clock)

def GetLogicClock() -> int:
    return _temp.LogicClock

def OpenUrl(url: str):
    sdl2.SDL_OpenURL(url)

def AllEvents() -> list:
    return _temp.Events

def QuitEvent() -> bool:
    for event in AllEvents():
        if event.type == 256:
            return True
    return False

def WindowEvents(window: int = All, event: int = All):
    for _event in AllEvents():
        if _event.type == 512:
            if window == _event.window.windowID or window == All:
                if event == _event.window.event or event == All:
                    return _event.window.event
    return 0

class AutoCalcs2D:
    @staticmethod
    def Percent(percent: int):
        return lambda location, size, size_window, scale: size_window / 100 * percent

    @staticmethod
    def Center():
        return lambda location, size, size_window, scale: size_window / 2

def _calculate_size(location, size, size_window, scale) -> list:
    return [round(size[num](location[num], size[num], size_window[num], scale[num]) * scale[num]) if callable(size[num]) else round(size[num] * scale[num]) for num in range(2)]

def _calculate_location(location, size, size_window, scale) -> list:
    return [round(location[num](location[num], size[num], size_window[num], scale[num])) if callable(location[num]) else round(location[num]) for num in range(2)]

def _calculate_object2d(location, size, rotation, scale, window, camera, pivot) -> list[bool, list[int, int], list[int, int]]:
    _location_camera = camera.location if camera is not None else window.camera.location
    _size_window = window.logicalResolution

    _size = _calculate_size((0, 0), size, _size_window, scale)
    _location = _calculate_location(location, _size, _size_window, scale)

    _location = [_location[0] + _location_camera[0], _location[1] + _location_camera[1]]

    if pivot == Pivot2D.Center:
        _location = [_location[0] - int(_size[0] // 2), _location[1] - int(_size[1] // 2)]
    elif pivot == Pivot2D.TopRightSide:
        _location[1] -= _size[1]
    elif pivot == Pivot2D.LowerLeftSide:
        _location[0] -= _size[0]
    elif pivot == Pivot2D.LowerRightSide:
        _location = [_location[0] - _size[0], _location[1] - _size[1]]

    render = False
    if _size[0] * -1 < _location[0] < _size_window[0] and _size[1] * -1 < _location[1] < _size_window[1]:
        render = True
    else:
        tt = edges(calculate_square_vertices(_location, _size, rotation))
        win_tt = edges([(0, 0), (0, window.resolution[1]), window.resolution, (window.resolution[0], 0)])

        for num in range(4):
            for num2 in range(4):
                if line_intersection(tt[num][0], tt[num][1], win_tt[num2][0], win_tt[num2][1]):
                    render = True

    return [render, _location, _size]

def _calculate_line(start, end, size, window, camera) -> list[bool, list[int, int], list[int, int], int]:
    _location_camera = camera.location if camera is not None else window.camera.location
    _size_window = window.resolution

    _size = _calculate_size((0, 0), (size, 0), _size_window, (1, 1))
    _start = _calculate_location(start, _size, _size_window, (1, 1))
    _end = _calculate_location(end, _size, _size_window, (1, 1))

    _start = [_start[0] + _location_camera[0], _start[1] + _location_camera[1]]
    _end = [_end[0] + _location_camera[0], _end[1] + _location_camera[1]]

    render = False
    if (_size[0] * -1 < _start[0] < _size_window[0] and (_size[0] * -1) < _start[1] < _size_window[1]) or ((_size[0] * -1) < _end[0] < _size_window[0] and (_size[0] * -1) < _end[1] < _size_window[1]):
        render = True
    else:
        win_tt = edges([(0, 0), (0, window.resolution[1]), window.resolution, (window.resolution[0], 0)])

        for num in range(4):
            if line_intersection(_start, _end, win_tt[num][0], win_tt[num][1]):
                render = True

    return [render, _start, _end, _size[0]]



from typing import NoReturn

__all__ = ["Log", "LogDebug", "LogInfo", "LogWarn", "LogError", "LogCritical"]

class _ConsoleColors:
    Reset = "\033[0m"
    Gray = "\033[90m"
    Red = "\033[91m"
    Green = "\033[92m"
    Yellow = "\033[93m"
    Blue = "\033[94m"
    Purple = "\033[95m"
    Cyan = "\033[96m"
    White = "\033[97m"

def Log(msg=""):
    print(f"{_ConsoleColors.Gray}{msg}{_ConsoleColors.Reset}")

def LogDebug(msg=""):
    print(f"{_ConsoleColors.Yellow}Debug{_ConsoleColors.Reset}: {msg}")

def LogInfo(msg=""):
    print(f"{_ConsoleColors.Green}Info{_ConsoleColors.Reset}: {_ConsoleColors.Gray}{msg}{_ConsoleColors.Reset}")

def LogWarn(msg=""):
    print(f"{_ConsoleColors.Yellow}Warning{_ConsoleColors.Reset}: {msg}")

def LogError(msg=""):
    print(f"{_ConsoleColors.Red}Error{_ConsoleColors.Reset}: {msg}")

def LogCritical(msg="", long_msg="", exit_code=1) -> NoReturn:
    print(f"{_ConsoleColors.Red}Critical Error{_ConsoleColors.Reset}: {msg}")
    if long_msg:
        Log(long_msg)
    raise SystemExit(exit_code)



from ..Log import *

LogCritical("system incompatible with 'MGE._InputsEmulator'", exit_code=0)

def MousePress(button):
    pass

def MouseRelease(button):
    pass

def MouseClick(button):
    pass

def MouseScroll(scroll):
    pass

def KeyboardPress(key):
    pass

def KeyboardRelease(key):
    pass

def KeyboardClick(key):
    pass



from numpy import deg2rad, cos, sin

__all__ = ["simpleMotion", "object2dSimpleMotion"]

def simpleMotion(location, theta, axis, speed=0):
    theta_rad = deg2rad((theta + 90 * axis) * -1)

    dx = speed * cos(theta_rad)
    dy = speed * sin(theta_rad)

    return [(location[0] + -dx), (location[1] + dy)]

def object2dSimpleMotion(object, axis, speed=0):
    if "speed" in object.variables:
        if object.variables["speed"] != 0:
            speed = object.variables["speed"]

    return simpleMotion(object.location, object.rotation, axis, speed)



import ctypes

user32 = ctypes.windll.user32

__all__ = ["MousePress", "MouseRelease", "MouseClick", "MouseScroll",
           "KeyboardPress", "KeyboardRelease", "KeyboardClick"]

_MouseEvent = {1: [0x0002, 0x0004], 2: [0x0020, 0x0040], 3: [0x0008, 0x0010], 4: [0x0002, 0x0004], 5: [0x0002, 0x0004]}
_KeyboardEvent = {}

def MousePress(button):
    button = 1 if button == -1 else button
    user32.mouse_event(_MouseEvent[button][0])

def MouseRelease(button):
    button = 1 if button == -1 else button
    user32.mouse_event(_MouseEvent[button][1])

def MouseClick(button):
    MousePress(button)
    MouseRelease(button)

def MouseScroll(scroll):
    user32.mouse_event(0x0800, 0, 0, scroll, 0)

def KeyboardPress(key):
    user32.keybd_event(key, user32.MapVirtualKeyA(key, 0), 0x0000 | 0x0001, 0)

def KeyboardRelease(key):
    user32.keybd_event(key, user32.MapVirtualKeyA(key, 0), 0x0002 | 0x0001, 0)

def KeyboardClick(key):
    KeyboardPress(key)
    KeyboardRelease(key)



from ctypes import Structure, c_int, c_char_p, POINTER as _P
from ._dll_loader import IMAGEFunc
from .sdl2 import SDL_version, SDL_RWops, SDL_Surface, SDL_Texture, SDL_Renderer

class IMG_Animation(Structure):
    _fields_ = [("w", c_int), ("h", c_int), ("count", c_int), ("frames", _P(_P(SDL_Surface))), ("delays", _P(c_int))]

def IMG_Linked_Version() -> SDL_version: return IMAGEFunc("IMG_Linked_Version", None, _P(SDL_version))()

def IMG_Init(flags=0) -> int: return IMAGEFunc("IMG_Init", [c_int], c_int)(flags)
def IMG_Quit(): IMAGEFunc("IMG_Quit")()

def IMG_Load(file: str) -> SDL_Surface: return IMAGEFunc("IMG_Load", [c_char_p], _P(SDL_Surface))(file.encode())
def IMG_Load_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_Load_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadTexture(renderer: SDL_Renderer, file: str) -> SDL_Texture: return IMAGEFunc("IMG_LoadTexture", [_P(SDL_Renderer), c_char_p], _P(SDL_Texture))(renderer, file.encode())

def IMG_isAVIF(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isAVIF", [_P(SDL_RWops)], c_int)(src))
def IMG_isICO(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isICO", [_P(SDL_RWops)], c_int)(src))
def IMG_isCUR(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isCUR", [_P(SDL_RWops)], c_int)(src))
def IMG_isBMP(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isBMP", [_P(SDL_RWops)], c_int)(src))
def IMG_isGIF(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isGIF", [_P(SDL_RWops)], c_int)(src))
def IMG_isJPG(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isJPG", [_P(SDL_RWops)], c_int)(src))
def IMG_isJXL(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isJXL", [_P(SDL_RWops)], c_int)(src))
def IMG_isLBM(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isLBM", [_P(SDL_RWops)], c_int)(src))
def IMG_isPCX(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isPCX", [_P(SDL_RWops)], c_int)(src))
def IMG_isPNG(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isPNG", [_P(SDL_RWops)], c_int)(src))
def IMG_isPNM(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isPNM", [_P(SDL_RWops)], c_int)(src))
def IMG_isSVG(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isSVG", [_P(SDL_RWops)], c_int)(src))
def IMG_isQOI(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isQOI", [_P(SDL_RWops)], c_int)(src))
def IMG_isTIF(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isTIF", [_P(SDL_RWops)], c_int)(src))
def IMG_isXCF(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isXCF", [_P(SDL_RWops)], c_int)(src))
def IMG_isXPM(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isXPM", [_P(SDL_RWops)], c_int)(src))
def IMG_isXV(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isXV", [_P(SDL_RWops)], c_int)(src))
def IMG_isWEBP(src: SDL_RWops) -> bool: return bool(IMAGEFunc("IMG_isWEBP", [_P(SDL_RWops)], c_int)(src))

def IMG_LoadAVIF_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadAVIF_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadICO_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadICO_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadCUR_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadCUR_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadBMP_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadBMP_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadGIF_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadGIF_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadJPG_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadJPG_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadJXL_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadJXL_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadLBM_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadLBM_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadPCX_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadPCX_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadPNG_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadPNG_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadPNM_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadPNM_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadSVG_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadSVG_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadQOI_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadQOI_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadTGA_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadTGA_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadTIF_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadTIF_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadXCF_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadXCF_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadXPM_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadXPM_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadXV_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadXV_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)
def IMG_LoadWEBP_RW(src: SDL_RWops) -> SDL_Surface: return IMAGEFunc("IMG_LoadWEBP_RW", [_P(SDL_RWops)], _P(SDL_Surface))(src)

def IMG_LoadSizedSVG_RW(src: SDL_RWops, width: int, height: int) -> SDL_Surface: return IMAGEFunc("IMG_LoadSizedSVG_RW", [_P(SDL_RWops), c_int, c_int], _P(SDL_Surface))(src, width, height)

def IMG_ReadXPMFromArray(xpm: str) -> SDL_Surface: return IMAGEFunc("IMG_ReadXPMFromArray", [_P(c_char_p)], _P(SDL_Surface))(xpm.encode())
def IMG_ReadXPMFromArrayToRGB888(xpm: str) -> SDL_Surface: return IMAGEFunc("IMG_ReadXPMFromArrayToRGB888", [_P(c_char_p)], _P(SDL_Surface))(xpm.encode())

def IMG_SavePNG(surface: SDL_Surface, file: str): return IMAGEFunc("IMG_SavePNG", [_P(SDL_Surface), c_char_p], c_int)(surface, file.encode())
def IMG_SaveJPG(surface: SDL_Surface, file: str, quality: int): return IMAGEFunc("IMG_SaveJPG", [_P(SDL_Surface), c_char_p, c_int], c_int)(surface, file.encode(), quality)
# NOTE: Not available in official macOS binaries

def IMG_LoadAnimation(file: str) -> IMG_Animation: return IMAGEFunc("IMG_LoadAnimation", [c_char_p], _P(IMG_Animation))(file.encode())
def IMG_LoadAnimation_RW(src: SDL_RWops) -> IMG_Animation: return IMAGEFunc("IMG_LoadAnimation_RW", [_P(SDL_RWops), c_int], _P(IMG_Animation))(src, 1)
def IMG_FreeAnimation(anim: IMG_Animation): IMAGEFunc("IMG_FreeAnimation", [_P(IMG_Animation)])(anim)



import os
from math import cos, sin, radians
from .Log import LogError

def line_intersection(start_1, end_1, start_2, end_2):
    dx1 = end_1[0] - start_1[0]
    dy1 = end_1[1] - start_1[1]
    dx2 = end_2[0] - start_2[0]
    dy2 = end_2[1] - start_2[1]

    det = dx1 * dy2 - dy1 * dx2

    if det == 0:
        return None

    u = ((start_2[0] - start_1[0]) * dy2 - (start_2[1] - start_1[1]) * dx2) / det
    v = ((start_2[0] - start_1[0]) * dy1 - (start_2[1] - start_1[1]) * dx1) / det

    if 0 <= u <= 1 and 0 <= v <= 1:
        intersection_x = start_1[0] + u * dx1
        intersection_y = start_1[1] + u * dy1
        return (round(intersection_x), round(intersection_y))
    else:
        return None

def rotate_point(x, y, cx, cy, angle):
    angle_rad = radians(angle)

    dx = x - cx
    dy = y - cy

    new_x = dx * cos(angle_rad) - dy * sin(angle_rad) + cx
    new_y = dx * sin(angle_rad) + dy * cos(angle_rad) + cy

    return new_x, new_y

def calculate_square_vertices(location, size, rotation_angle):
    x, y = location
    width, height = size

    cx = x + width / 2
    cy = y + height / 2

    vertices = [(x, y), (x + width, y), (x + width, y + height), (x, y + height)]

    if rotation_angle != 0:
        vertices = [rotate_point(v[0], v[1], cx, cy, rotation_angle) for v in vertices]

    return vertices

def edges(vertices):
    _edges = [(vertices[num], vertices[num+1 if num+1 <= len(vertices)-1 else 0]) for num in range(len(vertices))]
    return _edges

class Mesh2D:
    def __init__(self, vertices: list):
        self._vertices = vertices
        self._edges = []
        self._faces = []

    @property
    def vertices(self):
        return self._vertices

    @property
    def edges(self):
        return [(self._vertices[num], self._vertices[num+1 if num+1 <= len(self._vertices)-1 else 0]) for num in range(len(self._vertices))]

def LoadObj(path):
    vertices = []
    if os.path.exists(path):
        with open(path) as obj:
            r = obj.read()
            for rr in r.split("\n"):
                if len(rr) > 0 and rr[0] == "v":
                    r2 = rr.split()
                    vertices.append((int(float(r2[1]) * 100), int(float(r2[3] if len(r2) == 4 else r2[2]) * 100)))
    if len(vertices) == 0:
        LogError(f"{path}")
    return Mesh2D(vertices)

def CreateMeshPlane(size):
    return Mesh2D([(0, 0), (size[0], 0), (size[0], size[1]), (0, size[1])])



from numpy import frombuffer, uint8, mean
from ctypes import c_uint8
from .Texture import Texture
from .Color import Color
from .Constants import Colors
from ._sdl import sdl2

__all__ = ["Material", "DefaultMaterial", "ButtonDefaultMaterials"]

class Material:
    def __init__(self, texture: Texture | tuple = (), color: Color = Colors.StandardColor, alpha: int = 255):
        self._color = color

        self._alpha = alpha
        self._blurr = 0

        self._textures: list = []
        self._Cache_textures = []
        self.addTexture(texture)

        self._surface = sdl2.SDL_CreateRGBSurface(0, 16, 16, 32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000).contents
        sdl2.SDL_FillRect(self._surface, None, (self._color.b << 0) | (self._color.g << 8) | (self._color.r << 16) | (self.alpha << 24))
        self._surface_color = Color(self._color.RGBA)
        self._renderer = None

        self._render = True

    def __repr__(self):
        return f"<%s.%s color={self._color.RGBA} at 0x%X>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            id(self),
        )

    def render(self, size: tuple | list = None):
        if self._render:
            if self._renderer is not None:
                sdl2.SDL_DestroyRenderer(self._renderer)
                self._renderer = None
            sdl2.SDL_FreeSurface(self._surface)
            if len(self._textures) > 0:
                _size = max(self._textures, key=lambda _img: _img.image.size[0] * _img.image.size[1], default=None).image.size
                _size = _size if size is None else (_size if _size[0] * _size[1] > size[0] * size[1] else size)
                self._surface = sdl2.SDL_CreateRGBSurface(0, *_size, 32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000).contents
                self._renderer = sdl2.SDL_CreateSoftwareRenderer(self._surface)
                for tx in self._textures:
                    tx.render(self._renderer)
                self._render = False
                self.update()

                ## ----------------------------------------------- ##

                width, height = self._surface.w, self._surface.h
                depth = self._surface.format.contents.BytesPerPixel * 8

                pixel_buffer = (c_uint8 * (width * height * depth // 8)).from_address(self._surface.pixels)
                surface_array = frombuffer(pixel_buffer, dtype=uint8)

                pixels = surface_array.reshape((height, width, depth // 8))

                media_cor = mean(pixels, axis=(0, 1))

                self._surface_color.RGBA = [int(i) for i in list(media_cor.astype(int))]
                self._surface_color.a = 255

                del media_cor, pixels, surface_array, pixel_buffer, height, width, depth

                ## ----------------------------------------------- ##

            else:
                self._surface = sdl2.SDL_CreateRGBSurface(0, 16, 16, 32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000).contents
                sdl2.SDL_FillRect(self._surface, None, (self._color.b << 0) | (self._color.g << 8) | (self._color.r << 16) | (self.alpha << 24))
                self._render = False

                self._surface_color.RGBA = self._color.RGBA
                self._surface_color.a = self.alpha

            return self.surface

    def update(self):
        self.render()
        _textures = []
        for tx in self._textures:
            _textures.append(tx.tx(self._renderer))
        if self._Cache_textures != _textures:
            self._Cache_textures = _textures
            sdl2.SDL_RenderClear(self._renderer)
            for tx in _textures:
                sdl2.SDL_RenderCopy(self._renderer, tx, None, None)
            return True
        return False

    @property
    def surface(self):
        return self._surface

    @property
    def surfaceColor(self) -> Color:
        return self._surface_color

    @property
    def color(self) -> Color:
        return self._color

    @property
    def colorRGB(self):
        return self._color.RGB

    @property
    def colorRGBA(self):
        return self._color.RGBA

    @color.setter
    def color(self, color: Color):
        self._color = color

    @colorRGB.setter
    def colorRGB(self, color):
        self._color.RGB = color

    @colorRGBA.setter
    def colorRGBA(self, color):
        self._color.RGBA = color

    @property
    def alpha(self) -> int:
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: int):
        self._alpha = alpha

    @property
    def blurr(self) -> int:
        return self._blurr

    @blurr.setter
    def blurr(self, blurr: int):
        self._blurr = blurr

    def addTexture(self, texture: Texture | tuple):
        if isinstance(texture, Texture):
            if texture not in self._textures:
                self._textures.append(texture)
        elif isinstance(texture, tuple):
            for tx in texture:
                if isinstance(tx, Texture):
                    if tx not in self._textures:
                        self._textures.append(tx)

    def removeTexture(self, texture: Texture | tuple | int):
        if isinstance(texture, Texture):
            if texture in self._textures:
                self._textures.remove(texture)
        elif isinstance(texture, int):
            self._textures.pop(texture)
        elif isinstance(texture, tuple):
            for tx in texture:
                if isinstance(tx, Texture):
                    if tx in self._textures:
                        self._textures.remove(tx)

    @property
    def textures(self) -> list[Texture]:
        return self._textures

    @textures.setter
    def textures(self, textures):
        self._textures = textures

DefaultMaterial = Material()

ButtonDefaultMaterials = [Material(color=Color(40, 40, 40)), Material(color=Color(35, 35, 35))]



import os
from ctypes import c_uint8, c_uint32
from numpy import frombuffer, uint8, mean, ndarray
from urllib.request import urlopen
from threading import Thread

from ._sdl import sdl2, sdlimage
from .Log import LogError
from .Constants import ImageFormat, Colors
from .Color import Color

__all__ = ["Image", "Icon", "CreateAnimatedImage",
           "LoadImage", "LoadImageUrl", "LoadIcon",
           "SaveImagePNG", "SaveImageJPG",
           "NewImage",
           "DefaultIcon",
           "image_to_icon", "icon_to_image", "image_to_numpyArray", "compare_image"]

class Image:
    def __init__(self):
        self.images = []
        self.delays = []
        self._size = (0, 0)
        self._format = None

    def __repr__(self):
        return f"<%s.%s size=%dx%d at 0x%X>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self._size[0],
            self._size[1],
            id(self),
        )

    def fill(self, color: Color):
        sdl2.SDL_FillRect(self.images[0], None, sdl2.SDL_MapRGBA(self.images[0].format, *color.RGBA))

    @property
    def count(self):
        return len(self.images)

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size: list | tuple):
        _images = []
        for _image in self.images:
            new_surface = sdl2.SDL_CreateRGBSurfaceWithFormat(0, size[0], size[1], self._format[0], self._format[1])
            sdl2.SDL_BlitScaled(_image, None, new_surface, None)
            sdl2.SDL_FreeSurface(_image)
            _images.append(new_surface.contents)
        del self.images, self._size
        self.images = _images
        self._size = list(size)

    @property
    def format(self):
        return self._format

    @format.setter
    def format(self, _format):
        _images = []
        for _image in self.images:
            new_surface = sdl2.SDL_ConvertSurfaceFormat(_image, _format[1], 0)
            sdl2.SDL_FreeSurface(_image)
            _images.append(new_surface.contents)
        del self.images
        self.images = _images
        self._format = _format

    @property
    def color(self):
        pixel_buffer = (c_uint8 * (self._size[0] * self._size[1] * self._format[0] // 8)).from_address(self.images[0].pixels)
        surface_array = frombuffer(pixel_buffer, dtype=uint8)

        pixels = surface_array.reshape((self._size[1], self._size[0], self._format[0] // 8))

        # Calculando a média de cada canal de cor (R, G, B, A)
        media_cor = mean(pixels, axis=(0, 1))

        _color = Color(list(media_cor.astype(int)))
        _color.a = 255

        return _color

    @property
    def animated(self):
        return len(self.images) > 1

    def crop(self, location: list | tuple, size: list | tuple):
        _images = []
        for _image in self.images:
            new_surface = sdl2.SDL_CreateRGBSurfaceWithFormat(0, size[0], size[1], self._format[0], self._format[1])
            sdl2.SDL_BlitSurface(_image, sdl2.SDL_Rect(*location, *size), new_surface, None)
            sdl2.SDL_FreeSurface(_image)
            _images.append(new_surface.contents)
        del self.images
        self.images = _images
        self._size = list(size)

    def copy(self):
        _new_image = Image()
        for _image in self.images:
            _new_image.images.append(sdl2.SDL_DuplicateSurface(_image).contents)
        _new_image.delays = self.delays
        _new_image._size = self.size
        _new_image._format = self.format
        return _new_image

    def close(self):
        for _image in self.images:
            sdl2.SDL_FreeSurface(_image)
        del self

class Icon:
    def __init__(self):
        self.icon = None
        self._size = (0, 0)

    def __repr__(self):
        return f"<%s.%s at 0x%X>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            id(self),
        )

    @property
    def size(self):
        return self._size

    def close(self):
        if self.icon is not None:
            sdl2.SDL_FreeSurface(self.icon)

def LoadImage(path: str) -> Image:
    _Image = Image()
    if os.path.exists(path):
        if path[-4:] == ".gif":
            _Gif = sdlimage.IMG_LoadAnimation(path).contents
            for frame in range(_Gif.count):
                _Image.images.append(_Gif.frames[frame].contents)
                _Image.delays.append(float(_Gif.delays[frame]))
        else:
            _Image.images.append(sdlimage.IMG_Load(path).contents)
            _Image.delays.append(0)

    if len(_Image.images) == 0 or not _Image.images[0]:
        LogError(f"Unable to load image '{path}'")
        _Image.images = [sdl2.SDL_CreateRGBSurfaceWithFormat(0, 16, 16, ImageFormat.ARGB[0], ImageFormat.ARGB[1]).contents]
        _Image.delays = [0]

    _Image._size = (_Image.images[0].w, _Image.images[0].h)
    _Image._format = ImageFormat.ARGB
    return _Image

def _loadImageUrl(url: str):
    _Image = Image()
    try:
        with urlopen(url) as res:
            _chunkSize = -1
            image_bytes = bytes()
            while True:
                chunk = res.read(_chunkSize if _chunkSize > 0 else -1)
                image_bytes += chunk
                if not chunk or _chunkSize <= 0:
                    del chunk
                    break
                del chunk
    except Exception as e:
        pass
    else:
        sdl_image = sdl2.SDL_RWFromConstMem(image_bytes, len(image_bytes))
        if sdlimage.IMG_isGIF(sdl_image):
            _Gif = sdlimage.IMG_LoadAnimation_RW(sdl_image).contents
            for frame in range(_Gif.count):
                _Image.images.append(_Gif.frames[frame].contents)
                _Image.delays.append(float(_Gif.delays[frame]))
        else:
            _Image.images.append(sdlimage.IMG_Load_RW(sdl_image).contents)
            _Image.delays.append(0)
        sdl2.SDL_FreeRW(sdl_image)
        del image_bytes, sdl_image

    if len(_Image.images) == 0 or not _Image.images[0]:
        LogError(f"Unable to load image '{url}'")
        _Image.images = [sdl2.SDL_CreateRGBSurfaceWithFormat(0, 16, 16, ImageFormat.ARGB[0], ImageFormat.ARGB[1]).contents]
        _Image.delays = [0]

    _Image._size = (_Image.images[0].w, _Image.images[0].h)
    _Image._format = ImageFormat.ARGB

    return _Image

def LoadImageUrl(url: str | list | tuple, mult: bool = False) -> Image | tuple | None:
    if isinstance(url, str):
        return _loadImageUrl(url)
    elif isinstance(url, (list, tuple)):
        _image = []
        _threads = []

        def _load(__num):
            _image.insert(__num, _loadImageUrl(url[__num]))

        for num in range(len(url)):
            if mult:
                _thread = Thread(target=_load, args=(num, ), daemon=True)
                _thread.start()
                _threads.append(_thread)
            else:
                _load(num)

        for t in _threads:
            t.join()

        return tuple(_image)
    return None

def LoadIcon(path: str) -> Icon:
    _Icon = Icon()
    if os.path.exists(path):
        if path[-4:] == ".ico":
            _Icon.icon = sdlimage.IMG_LoadICO_RW(sdl2.SDL_RWFromFile(path, "r")).contents
    if not _Icon.icon:
        LogError(f"Unable to load icon '{path}'")
        _Icon.icon = sdl2.SDL_CreateRGBSurfaceWithFormat(0, 16, 16, ImageFormat.ARGB[0], ImageFormat.ARGB[1]).contents
    _Icon._size = (_Icon.icon.w, _Icon.icon.h)
    return _Icon

def NewImage(mode=ImageFormat.ARGB, size: list | tuple = (32, 32), color: Color = Colors.StandardColor) -> Image:
    _image = Image()
    _image.images.append(sdl2.SDL_CreateRGBSurfaceWithFormat(0, size[0], size[1], mode[0], mode[1]).contents)
    _image._size = list(size)
    _image._format = mode
    _image.delays.append(0)
    if color is not None:
        _image.fill(color)
    return _image

def CreateAnimatedImage(images: list[Image] | tuple[Image]):
    _Image = Image()
    _Image._size = max(images, key=lambda _img: _img.size[0] * _img.size[1], default=None).size
    _Image._format = ImageFormat.ARGB
    for img in images:
        for _image in img.images:
            if _Image.size[0] > img.size[0] or _Image.format != img.format:
                new_surface = sdl2.SDL_CreateRGBSurfaceWithFormat(0, _Image.size[0], _Image.size[1], _Image.format[0], _Image.format[1])
                sdl2.SDL_BlitScaled(_image, None, new_surface, None)
                _image = new_surface.contents
            _Image.images.append(_image)
        for _delay in img.delays:
            _Image.delays.append(_delay)

    return _Image

def SaveImagePNG(image: Image, path: str = "image.png"):
    sdlimage.IMG_SavePNG(image.images[0], path)

def SaveImageJPG(image: Image, path: str = "image.jpg", quality: int = 80):
    sdlimage.IMG_SaveJPG(image.images[0], path, quality)

#def SaveImageGIF(image: Image, path="image.GIF"):
#    pass

def image_to_icon(image: Image) -> Icon:
    icon = Icon()
    icon.icon = sdl2.SDL_ConvertSurfaceFormat(image.images[0], ImageFormat.ARGB[1], 0).contents
    icon._size = image.size
    return icon

def icon_to_image(icon: Icon) -> Image:
    _image = Image()
    _image.images.append(sdl2.SDL_DuplicateSurface(icon.icon).contents)
    _image.delays.append(0)
    _image._size = icon.size
    _image._format = ImageFormat.ARGB
    return _image

def _sdl2_surface_to_numpy_array(surface: sdl2.SDL_Surface) -> ndarray:
    width, height = surface.w, surface.h
    depth = surface.format.contents.BytesPerPixel * 8

    pixel_buffer = (c_uint8 * (width * height * depth // 8)).from_address(surface.pixels)
    surface_array = frombuffer(pixel_buffer, dtype=uint8)
    surface_array = surface_array.reshape((height, width, depth // 8))

    return surface_array.copy()

def image_to_numpyArray(image: Image) -> ndarray:
    return _sdl2_surface_to_numpy_array(image.images[0])

def compare_image(image1: Image, image2: Image) -> bool:
    if image1.size != image2.size:
        return False
    pixels1 = image_to_numpyArray(image1)
    pixels2 = image_to_numpyArray(image2)
    pixels = pixels1 == pixels2
    return bool(pixels.all())

def _test_RGBA_mod(surface1, surface2, mult, r, g, b, a):
    if surface1.w == surface2.w and surface1.h == surface2.h:
        pixels, pitch = surface1.pixels, surface1.pitch
        bytes_per_pixel = surface1.format.contents.BytesPerPixel

        pixels2 = surface2.pixels

        def _line_mod():
            for x in range(surface1.w):
                pixel_offset = y * pitch + x * bytes_per_pixel

                _a = c_uint8.from_address((pixels if a else pixels2) + pixel_offset + 3).value
                _r = c_uint8.from_address((pixels if r else pixels2) + pixel_offset + 0).value
                _g = c_uint8.from_address((pixels if g else pixels2) + pixel_offset + 1).value
                _b = c_uint8.from_address((pixels if b else pixels2) + pixel_offset + 2).value

                pixel_value = (_a << 0) | (_r << 8) | (_g << 16) | (_b << 24)

                c_uint32.from_address(surface2.pixels + y * surface2.pitch + x * 4).value = pixel_value

        _threads = []

        for y in range(surface1.h):
            if mult:
                _thread = Thread(target=_line_mod, daemon=True)
                _thread.start()
                _threads.append(_thread)
            else:
                _line_mod()

        for t in _threads:
            t.join()

_path = f"{os.path.abspath(os.path.dirname(__file__))}/logo.ico"
if os.path.exists("./logo.ico"):
    DefaultIcon = LoadIcon("./logo.ico")
elif os.path.exists(_path):
    DefaultIcon = LoadIcon(_path)
else:
    _icon = NewImage(ImageFormat.ARGB, (1, 1), Color((0, 0, 0)))
    DefaultIcon = image_to_icon(_icon)
    _icon.close()



from .Log import LogCritical
from ._sdl import sdl2

__all__ = ("Monitors", "Monitor", "_update_monitors_datas")

class Monitor:
    def __init__(self, index: int, name: str, resolution: tuple[int, int], frame_rate: float):
        self.__Monitor_Id__ = index
        self.name = name
        self.resolution = resolution
        self.frameRate = frame_rate

Monitors: list[Monitor] = []

def _update_monitors_datas():
    ret = sdl2.SDL_GetNumVideoDisplays()
    if ret < 0:
        LogCritical("no monitor found")
    for num in range(ret):
        name = sdl2.SDL_GetDisplayName(num)
        display_mode = sdl2.SDL_GetCurrentDisplayMode(num)
        Monitors.append(Monitor(num, name, (display_mode.w, display_mode.h), display_mode.refresh_rate))



from ..Platform import Platform

__all__ = ["version",
           "MousePress", "MouseRelease", "MouseClick", "MouseScroll",
           "KeyboardPress", "KeyboardRelease", "KeyboardClick"]

if Platform.system == "Windows":
    from ._win32 import *
#elif Platform.system == "Linux":
#    from ._linux import *
else:
    from ._error import *

version = "0.1.0"



from .Window import Window
from .Camera import Camera
from .Material import Material
from .Common import _calculate_line

__all__ = ["Line"]

class Line:
    def __init__(self, start, end, size: int):
        self._start = start
        self._end = end
        self._size = size

        self._material = Material()

    def drawObject(self, window: Window, camera: Camera = None):
        if window.__WindowActive__:
            if self not in window.drawnObjects:
                window.drawnObjects.append(self)
                render, cache_start, cache_end, cache_size = _calculate_line(self._start, self._end, self._size, window, camera)
                if render:
                    window.drawLine(cache_start, cache_end, cache_size, self._material.color)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        self._end = end

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size: int):
        self._size = int(size)

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material: Material):
        if self._material != material:
            self._material = material



from ctypes import c_int, c_uint8, byref, sizeof, windll

from .Common import AllEvents, WindowEvents, _temp, _calculate_object2d
from .Keyboard import KeyboardState
from .Time import Time, fps_to_time, get_fps_from_time
from ._sdl import sdl2, sdlgfx
from .Color import Color
from .Texture import _get_sdl2_texture_size
from .Camera import Camera
from .Image import Image, Icon, DefaultIcon, image_to_icon
from .Platform import Platform
from .Monitors import Monitors
from .Constants import All, WindowFlag, ImageFormat, Pivot2D
from .Mesh import rotate_point
from .Log import LogError, LogCritical
from .Mouse import object2dSimpleHover

__all__ = ["Window", "CreateGlWindow", "InternalWindow"]

class _Window:
    def __init__(self):
        self.__WindowActive__ = True

        self._resolution = None

        self.window = None
        self.context = None
        self.renderer = None
        self.__WindowId__ = -1

        self._variables = {}

        self._frameRateLimit = Monitors[0].frameRate
        self._cache = {"clear_screen": True, "fill": True, "times": {"standard_time": Time(fps_to_time(self._frameRateLimit)), "title_time": Time(0.5), "optimized_time": {"time_start": Time(0.5), "time_loop": Time(fps_to_time(2))}}, "window_tx": None}
        self.frameRate = 0

        self.drawnObjects = []
        self.render_all_objects = True

    @property
    def flags(self):
        if not self.__WindowId__ < 0:
            return sdl2.SDL_GetWindowFlags(self.window)
        else:
            return 0

    @property
    def frameRateLimit(self):
        return self._frameRateLimit

    @frameRateLimit.setter
    def frameRateLimit(self, value):
        self._frameRateLimit = value
        self._cache["times"]["standard_time"].delta_time = fps_to_time(self._frameRateLimit) if self._frameRateLimit > 0 else 0

    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, variables):
        self._variables = variables

    def close(self):
        self.clear(True)
        if self.window is not None:
            if self.__WindowId__ < 0:
                sdl2.SDL_FreeSurface(self.window)
            else:
                sdl2.SDL_DestroyWindow(self.window)
        if self.renderer is not None:
            sdl2.SDL_DestroyRenderer(self.renderer)
        if self.context is not None:
            sdl2.SDL_GL_DeleteContext(self.context)
        if self.__WindowId__ < 0 and self._cache["window_tx"] is not None:
            sdl2.SDL_DestroyTexture(self._cache["window_tx"])
            self._cache["window_tx"] = None
        self.window = self.renderer = None
        self.__WindowActive__ = False

    def clear(self, force=False, color=(0, 0, 0, 255)):
        if self.__WindowActive__ and not (self.flags >> 1) & 1:
            if force or self._cache["clear_screen"]:
                tmp = self.color
                self.color = color
                sdl2.SDL_RenderClear(self.renderer)
                self.color = tmp
                self._cache["clear_screen"] = False
                self._cache["fill"] = True

    def getImage(self):
        _Image = Image()
        if self.__WindowId__ < 0:
            _Image.images.append(self.window)
        else:
            _Image.images.append(sdl2.SDL_CreateRGBSurface(0, self._resolution[0], self._resolution[1], 32, 0, 0, 0, 0).contents)
            sdl2.SDL_RenderReadPixels(self.renderer, None, sdl2.SDL_PIXELFORMAT_ARGB8888, _Image.images[0].pixels, _Image.images[0].pitch)
        _Image.delays.append(0)
        _Image._size = self._resolution
        _Image._format = ImageFormat.ARGB
        return _Image

    def blit(self, surface: Image | sdl2.SDL_Surface | sdl2.SDL_Texture, location=(0, 0), size=None, rotation: int = 0):
        if self.__WindowActive__ and not (self.flags >> 1) & 1:
            create_texture = False
            if isinstance(surface, Image):
                texture = sdl2.SDL_CreateTextureFromSurface(self.renderer, surface.images[0])
                sdl2.SDL_SetTextureScaleMode(texture, 1)
                create_texture = True
                _size = surface.size if size is None else size
            elif isinstance(surface, sdl2.SDL_Surface):
                texture = sdl2.SDL_CreateTextureFromSurface(self.renderer, surface)
                sdl2.SDL_SetTextureScaleMode(texture, 1)
                create_texture = True
                _size = (surface.w, surface.h) if size is None else size
            elif isinstance(surface, sdl2.SDL_Texture):
                texture = surface
                _size = _get_sdl2_texture_size(texture) if size is None else size
            else:
                return

            if rotation != 0:
                sdl2.SDL_RenderCopyExF(self.renderer, texture, None, sdl2.SDL_FRect(*location, *_size), rotation, sdl2.SDL_FPoint(_size[0] // 2, _size[1] // 2), sdl2.SDL_FLIP_NONE)
            else:
                sdl2.SDL_RenderCopy(self.renderer, texture, None, sdl2.SDL_Rect(*location, *_size))
            if create_texture:
                sdl2.SDL_DestroyTexture(texture)

    def drawPixel(self, location, color: Color):
        sdlgfx.pixelRGBA(self.renderer, *location, *color.RGBA)

    def drawSquare(self, location, size, rotation, radius, color: Color):
        if self.__WindowActive__ and not (self.flags >> 1) & 1:
            def _square(renderer_, location_, size_, radius_, color_):
                if isinstance(radius_, (tuple, list)) and len(radius_) == 4:
                    if all(x == radius_[0] for x in radius_):
                        sdlgfx.roundedBoxRGBA(renderer_, location_[0], location_[1], location_[0] + size_[0] - 1, location_[1] + size_[1] - 1, radius_[0], *color_.RGBA)
                    else:
                        sdlgfx.filledPieRGBA(renderer_, location_[0] + radius_[0], location_[1] + radius_[0], radius_[0], 180, 270, *color_.RGBA)
                        sdlgfx.filledPieRGBA(renderer_, location_[0] + radius_[1] - 1, location_[1] + size_[1] - radius_[1] - 1, radius_[1], 90, 180, *color_.RGBA)
                        sdlgfx.filledPieRGBA(renderer_, location_[0] + radius_[1] + size_[0] - radius_[1] - radius_[2] - 1, location_[1] + size_[1] - radius_[2] - 1, radius_[2], 0, 90, *color_.RGBA)
                        sdlgfx.filledPieRGBA(renderer_, location_[0] + radius_[0] + size_[0] - radius_[0] - radius_[3] - 1, location_[1] + radius_[3], radius_[3], 270, 360, *color_.RGBA)
                        sdlgfx.boxRGBA(renderer_, location_[0] + radius_[0], location_[1], location_[0] + size_[0] - 1 - radius_[3], location_[1] + size_[1] // 2 - 1, *color_.RGBA)
                        sdlgfx.boxRGBA(renderer_, location_[0] + radius_[1], location_[1] + size_[1] // 2, location_[0] + size_[0] - 1 - radius_[2], location_[1] + size_[1] - 1, *color_.RGBA)
                        sdlgfx.boxRGBA(renderer_, location_[0], location_[1] + radius_[0], location_[0] + size_[0] // 2 - 1, location_[1] + size_[1] - 1 - radius_[1], *color_.RGBA)
                        sdlgfx.boxRGBA(renderer_, location_[0] + size_[0] // 2, location_[1] + radius_[3], location_[0] + size_[0] - 1, location_[1] + size_[1] - 1 - radius_[2], *color_.RGBA)
                elif isinstance(radius_, int) and radius_ > 0:
                    sdlgfx.roundedBoxRGBA(renderer_, location_[0], location_[1], location_[0] + size_[0] - 1, location_[1] + size_[1] - 1, radius_, *color_.RGBA)
                else:
                    sdlgfx.boxRGBA(renderer_, location_[0], location_[1], location_[0] + size_[0] - 1, location_[1] + size_[1] - 1, *color_.RGBA)
            if rotation != 0:
                mask = sdl2.SDL_CreateRGBSurface(0, size[0], size[1], 32, 0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000).contents
                _renderer = sdl2.SDL_CreateSoftwareRenderer(mask)
                _square(_renderer, [0, 0], size, radius, color)
                _texture = sdl2.SDL_CreateTextureFromSurface(self.renderer, mask).contents
                sdl2.SDL_SetTextureScaleMode(_texture, 1)
                self.blit(_texture, [location[0], location[1]], size, rotation)
                sdl2.SDL_DestroyTexture(_texture)
                sdl2.SDL_DestroyRenderer(_renderer)
                sdl2.SDL_FreeSurface(mask)
            else:
                _square(self.renderer, location, size, radius, color)

    def drawEdgesSquare(self, location, size, rotation, line_size, radius, color: Color):
        if self.__WindowActive__ and not (self.flags >> 1) & 1:
            if line_size != 0:
                def _hollow_square(renderer_, location_, size_, line_size_, radius_, color_):
                    for num in range(line_size_ if line_size_ > 0 else -line_size_):
                        num = num if line_size_ > 0 else -num
                        if radius_ > 0:
                            sdlgfx.roundedRectangleRGBA(renderer_, location_[0] - num, location_[1] - num, location_[0] + size_[0] + num, location_[1] + size_[1] + num, radius_, *color_.RGBA)
                        else:
                            sdlgfx.rectangleRGBA(renderer_, location_[0] - num, location_[1] - num, location_[0] + size_[0] + num, location_[1] + size_[1] + num, *color_.RGBA)
                if rotation not in [0, 90, 180, 240, 360]:
                    _size = [size[0] + (line_size * 2 if line_size > 1 else 0), size[1] + (line_size * 2 if line_size > 1 else 0)]
                    _location = [line_size if line_size > 1 else 0, line_size if line_size > 1 else 0]
                    mask = sdl2.SDL_CreateRGBSurface(0, _size[0], _size[1], 32, 0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000).contents
                    _renderer = sdl2.SDL_CreateSoftwareRenderer(mask)
                    _hollow_square(_renderer, _location, size, line_size, radius, color)
                    self.blit(mask, [location[0] - _location[0], location[1] - _location[1]], None, rotation)
                    sdl2.SDL_DestroyRenderer(_renderer)
                    sdl2.SDL_FreeSurface(mask)
                else:
                    _hollow_square(self.renderer, location, size, line_size, radius, color)

    def drawLine(self, start, end, size, color: Color):
        if self.__WindowActive__ and not (self.flags >> 1) & 1:
            if size != 0:
                if size == 1:
                    sdlgfx.aalineRGBA(self.renderer, *start, *end, *color.RGBA)
                else:
                    sdlgfx.thickLineRGBA(self.renderer, *start, *end, size if size > 0 else -size, *color.RGBA)

    def drawPolygon(self, location, scale, rotation: int, mesh: list[list[int, int]], color: Color):
        if self.__WindowActive__ and not (self.flags >> 1) & 1:
            _mesh = [(round(point[0] * scale[0]), round(point[1] * scale[1])) for point in mesh]

            if rotation != 0:
                _max_vx = max(round(point[0] * scale[0]) for point in mesh)
                _max_vy = max(round(point[1] * scale[1]) for point in mesh)
                _mesh = [rotate_point(v[0], v[1], _max_vx / 2, _max_vy / 2, rotation) for v in _mesh]

            _vx = [point[0] + location[0] for point in _mesh]
            _vy = [point[1] + location[1] for point in _mesh]
            _n = len(_vx)

            sdlgfx.filledPolygonRGBA(self.renderer, (sdl2.Sint16 * _n)(*map(int, _vx)), (sdl2.Sint16 * _n)(*map(int, _vy)), _n, *color.RGBA)

    def drawEdgesPolygon(self, location, scale, rotation, mesh, color: Color):
        if self.__WindowActive__ and not (self.flags >> 1) & 1:
            _mesh = [(round(point[0] * scale[0]), round(point[1] * scale[1])) for point in mesh]

            if rotation != 0:
                _max_vx = max(round(point[0] * scale[0]) for point in mesh)
                _max_vy = max(round(point[1] * scale[1]) for point in mesh)
                _mesh = [rotate_point(v[0], v[1], _max_vx / 2, _max_vy / 2, rotation) for v in _mesh]

            _vx = [point[0] + location[0] for point in _mesh]
            _vy = [point[1] + location[1] for point in _mesh]
            _n = len(_vx)

            sdlgfx.polygonRGBA(self.renderer, (sdl2.Sint16 * _n)(*map(int, _vx)), (sdl2.Sint16 * _n)(*map(int, _vy)), _n, *color.RGBA)

    @property
    def color(self) -> tuple:
        if not (self.flags >> 1) & 1:
            r, g, b, a = c_uint8(0), c_uint8(0), c_uint8(0), c_uint8(0)
            ret = sdl2.SDL_GetRenderDrawColor(self.renderer, byref(r), byref(g), byref(b), byref(a))
            if ret is None:
                #sys.exit("Window_error")
                LogError("error getting render color")
            return r.value, g.value, b.value, a.value
        return 0, 0, 0, 0

    @color.setter
    def color(self, color: Color | list[int, int, int, int]):
        if not (self.flags >> 1) & 1:
            ret = sdl2.SDL_SetRenderDrawColor(self.renderer, *color)
            if ret < 0:
                LogError("error when changing rendering color")

class Window(_Window):
    def __init__(self, title="MGE", icon: Icon = DefaultIcon, resolution=(1280, 720), location=(300, 300), logical_resolution=(0, 0), shape=None, camera=Camera(), render_driver=All, flags=WindowFlag.Shown | WindowFlag.Resizable):
        super().__init__()

        self._title = title
        self._location = location
        self._resolution = resolution
        self._logical_resolution = logical_resolution

        self.camera = camera

        self._shape = shape
        if self._shape is None:
            self.window = sdl2.SDL_CreateWindow(self._title, self._location[0], self._location[1], self._resolution[0], self._resolution[1], flags)
        else:
            self.window = sdl2.SDL_CreateShapedWindow(self._title.encode(), self._location[0], self._location[1], self._resolution[0], self._resolution[1], flags)
            sdl2.SDL_SetWindowShape(self.window, self._shape, sdl2.SDL_WindowShapeMode())
        if (flags >> 1) & 1:
            self.context = sdl2.SDL_GL_CreateContext(self.window)
        else:
            self.renderer = sdl2.SDL_CreateRenderer(self.window, render_driver, 0x00000002 | 0x00000008 if Platform.drivers[render_driver].hardware or render_driver == -1 else 0x00000001 | 0x00000008)
        self.__WindowId__ = sdl2.SDL_GetWindowID(self.window)
        _wmInfo = sdl2.SDL_SysWMinfo()
        sdl2.SDL_GetWindowWMInfo(self.window, _wmInfo)
        self._hwnd = _wmInfo.info.win.window
        del _wmInfo
        self.logicalResolution = self._logical_resolution
        self.icon = self._icon = icon

    def __repr__(self):
        return f"<%s.%s resolution=%dx%d at 0x%X>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self._resolution[0],
            self._resolution[1],
            id(self)
        )

    @property
    def hwnd(self):
        return self._hwnd

    def update(self, still_frame_optimization: bool = False):
        if self.__WindowActive__:
            if still_frame_optimization:
                if AllEvents() or any(KeyboardState()):
                    self._cache["times"]["optimized_time"]["time_start"].restart()
                else:
                    if self._cache["times"]["optimized_time"]["time_start"].tick():
                        if not self._cache["times"]["optimized_time"]["time_loop"].tick(True):
                            return False
            if self._cache["times"]["standard_time"].tick(True):
                self.frameRate = get_fps_from_time(self._cache["times"]["standard_time"])

                self._resolution = list(sdl2.SDL_GetWindowSize(self.window))

                if WindowEvents(self.__WindowId__, 14):
                    self.close()
                    return

                self.drawnObjects.clear()
                self._cache["clear_screen"] = True
                self._cache["fill"] = True

                if (self.flags >> 1) & 1:
                    sdl2.SDL_GL_SwapWindow(self.window)
                else:
                    sdl2.SDL_RenderPresent(self.renderer)
                return True
        return False

    def recreate(self, title="MGE", icon: Icon = DefaultIcon, resolution=(1280, 720), location=(300, 300), logical_resolution=(0, 0), shape=None, camera=Camera(), flags=WindowFlag.Shown | WindowFlag.Resizable):
        self.close()
        self.__init__(title=title, icon=icon, resolution=resolution, location=location, logical_resolution=logical_resolution, shape=shape, camera=camera, flags=flags)

    def restore(self):
        sdl2.SDL_RestoreWindow(self.window)

    def show(self):
        sdl2.SDL_ShowWindow(self.window)

    def hide(self):
        sdl2.SDL_HideWindow(self.window)

    def maximize(self):
        sdl2.SDL_MaximizeWindow(self.window)

    def minimize(self):
        sdl2.SDL_MinimizeWindow(self.window)

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title: str):
        if self._title != title:
            if self._cache["times"]["title_time"].tick(True):
                self._title = title
                sdl2.SDL_SetWindowTitle(self.window, title)

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, icon: Image | Icon):
        if isinstance(icon, Icon):
            icon = icon
        elif isinstance(icon, Image):
            icon = image_to_icon(icon)
        else:
            LogError("incompatible icon")
            return
        sdl2.SDL_SetWindowIcon(self.window, icon.icon)
        self._icon = icon

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        if self._shape != shape:
            self._shape = shape
            if sdl2.SDL_IsShapedWindow(self.window):
                sdl2.SDL_SetWindowShape(self.window, self._shape, sdl2.SDL_WindowShapeMode())
            else:
                self.recreate(title=self._title, icon=self._icon, resolution=self._resolution, location=self._location, shape=self._shape, camera=self.camera, flags=self.flags)

    @property
    def opacity(self):
        return sdl2.SDL_GetWindowOpacity(self.window)

    @opacity.setter
    def opacity(self, opacity):
        sdl2.SDL_SetWindowOpacity(self.window, opacity)

    @property
    def location(self):
        if self.__WindowActive__:
            x, y = c_int(), c_int()
            sdl2.SDL_GetWindowPosition(self.window, x, y)
            return x.value, y.value
        return self._location

    @location.setter
    def location(self, location):
        if self.__WindowActive__:
            sdl2.SDL_SetWindowPosition(self.window, location[0], location[1])
        self._location = location[0], location[1]

    @property
    def resolution(self) -> tuple:
        if self.__WindowActive__:
            return sdl2.SDL_GetWindowSize(self.window)
        return self._resolution

    @resolution.setter
    def resolution(self, resolution):
        if self.__WindowActive__:
            sdl2.SDL_SetWindowSize(self.window, resolution[0], resolution[1])
        self._resolution = resolution

    @property
    def minimumResolution(self):
        w, h = c_int(), c_int()
        sdl2.SDL_GetWindowMinimumSize(self.window, w, h)
        return w.value, h.value

    @minimumResolution.setter
    def minimumResolution(self, resolution):
        sdl2.SDL_SetWindowMinimumSize(self.window, resolution[0], resolution[1])

    @property
    def maximumResolution(self):
        w, h = c_int(), c_int()
        sdl2.SDL_GetWindowMaximumSize(self.window, w, h)
        return w.value, h.value

    @maximumResolution.setter
    def maximumResolution(self, resolution):
        sdl2.SDL_SetWindowMaximumSize(self.window, resolution[0], resolution[1])

    @property
    def logicalResolution(self) -> tuple:
        if not (self.flags >> 1) & 1:
            r1, r2 = c_int(), c_int()
            sdl2.SDL_RenderGetLogicalSize(self.renderer, r1, r2)
            self._logical_resolution = [r1.value, r2.value]
        return self._logical_resolution if self._logical_resolution[0] != 0 and self._logical_resolution[1] != 0 else self.resolution

    @logicalResolution.setter
    def logicalResolution(self, resolution):
        if not (self.flags >> 1) & 1:
            sdl2.SDL_RenderSetLogicalSize(self.renderer, resolution[0], resolution[1])
            self._logical_resolution = resolution

    @property
    def borderless(self):
        return bool((self.flags >> 4) & 1)

    @borderless.setter
    def borderless(self, v: bool):
        sdl2.SDL_SetWindowBordered(self.window, v)

    @property
    def resizable(self):
        return bool((self.flags >> 5) & 1)

    @resizable.setter
    def resizable(self, v: bool):
        sdl2.SDL_SetWindowResizable(self.window, v)

    @property
    def alwaysOnTop(self):
        return bool((self.flags >> 15) & 1)

    @alwaysOnTop.setter
    def alwaysOnTop(self, v: bool):
        sdl2.SDL_SetWindowAlwaysOnTop(self.window, v)

    def set_TitleBarColor(self, color: Color):
        if Platform.system.lower() == "windows" and Platform.system_version >= 22000:
            windll.dwmapi.DwmSetWindowAttribute(self._hwnd, 35, byref(c_int(color.Uint32 & ~0xFF000000)), sizeof(c_int))
        else:
            LogError("")

    def set_BorderColor(self, color: Color | None):
        if Platform.system.lower() == "windows" and Platform.system_version >= 22000:
            windll.dwmapi.DwmSetWindowAttribute(self._hwnd, 34, byref(c_int(0xFFFFFFFE if color is None or color.a == 0 else color.Uint32 & ~0xFF000000)), sizeof(c_int))
        else:
            LogError("")

    def set_TitleColor(self, color: Color):
        if Platform.system.lower() == "windows" and Platform.system_version >= 22000:
            windll.dwmapi.DwmSetWindowAttribute(self._hwnd, 36, byref(c_int(color.Uint32 & ~0xFF000000)), sizeof(c_int))
        else:
            LogError("")

def CreateGlWindow(title="MGE", icon: Icon = DefaultIcon, resolution=(1280, 720), location=(300, 300), shape=None, msaa=0, flags=WindowFlag.Shown | WindowFlag.Resizable) -> Window:
    if msaa:
        sdl2.SDL_GL_SetAttribute(sdl2.SDL_GL_MULTISAMPLESAMPLES, msaa)
    _window = Window(title=title, icon=icon, resolution=resolution, location=location, shape=shape, flags=flags | WindowFlag.Opengl)
    return _window

class InternalWindow(_Window):
    def __init__(self, location=(0, 0), rotation: int = 0, size=(1280, 720), scale=(1, 1), resolution=(1280, 720), camera=Camera()):
        super().__init__()

        self._size = list(size)
        self._location = list(location)
        self._rotation = rotation
        self._scale = scale
        self._resolution = resolution
        self._pivot = Pivot2D.TopLeftSide
        self._border_size = 0
        self._border_color = Color((100, 100, 255))
        self._border_radius = [0, 0, 0, 0]
        self._cursor = 11

        self.camera = camera

        self.window = sdl2.SDL_CreateRGBSurface(0, self._resolution[0], self._resolution[1], 32, 0, 0, 0, 0).contents
        self.renderer = sdl2.SDL_CreateSoftwareRenderer(self.window)

        self._clear_color = (0, 0, 0, 255)

        self._frameRateLimit = 60

        self.object_render = False

    def __repr__(self):
        return f"<%s.%s resolution=%dx%d at 0x%X>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self._resolution[0],
            self._resolution[1],
            id(self)
        )

    def render(self, window=None):
        if self._cache["window_tx"] is not None:
            sdl2.SDL_DestroyTexture(self._cache["window_tx"])
            self._cache["window_tx"] = None
        self._cache["window_tx"] = sdl2.SDL_CreateTextureFromSurface(window.renderer if window is not None else sdl2.SDL_GetRenderer(sdl2.SDL_GetWindowFromID(1)), self.window).contents
        self.object_render = True

    def drawObject(self, window, camera: Camera = None, still_frame_optimization: bool = False):
        if window.__WindowActive__ and self.__WindowActive__:
            if self not in window.drawnObjects:
                window.drawnObjects.append(self)
                render, cache_location, cache_size = _calculate_object2d(self._location, self._size, self._rotation, self._scale, window, camera, self._pivot)
                if render:
                    if len(self.drawnObjects) == 0:
                        window.drawSquare(cache_location, cache_size, self._rotation, 0, Color(self._clear_color))
                        self.frameRate = self._frameRateLimit
                        return

                    if still_frame_optimization:
                        if AllEvents() or any(_temp.KeyboardState):
                            self._cache["times"]["optimized_time"]["time_start"].restart()
                        else:
                            if self._cache["times"]["optimized_time"]["time_start"].tick():
                                if not self._cache["times"]["optimized_time"]["time_loop"].tick(True):
                                    window.blit(self._cache["window_tx"], cache_location, cache_size, 0)
                                    return
                    if self._cache["times"]["standard_time"].tick(True):
                        self.frameRate = get_fps_from_time(self._cache["times"]["standard_time"])

                        self.drawnObjects.clear()
                        self._cache["clear_screen"] = True
                        self._cache["fill"] = True

                        if not self.object_render:
                            self.render()
                        else:
                            sdl2.SDL_UpdateTexture(self._cache["window_tx"], None, self.window.pixels, self.window.pitch)

                    window.blit(self._cache["window_tx"], cache_location, cache_size, self._rotation)

    def hover(self, window, camera: Camera = None) -> bool:
        if object2dSimpleHover(window, camera, self._location, self._size, self._scale, self._pivot):
            #_temp.MouseCursor = self._cursor
            return True
        return False

    def recreate(self, location=(0, 0), rotation: int = 0, size=(1280, 720), scale=(1, 1), resolution=(1280, 720), camera=Camera()):
        self.close()
        self.__init__(location=location, rotation=rotation, size=size, scale=scale, resolution=resolution, camera=camera)

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        self._location = location[0], location[1]

    @property
    def size(self) -> list[int, int]:
        # return calculate_size(self.size, Program.screen.get_size())
        return self._size

    @size.setter
    def size(self, size):
        self._size = size

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, rotation: int | float):
        if self._rotation != rotation:
            self._rotation = round(rotation, 4)
            self._rotation %= 360

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale):
        self._scale = scale

    @property
    def pivot(self):
        return self._pivot

    @pivot.setter
    def pivot(self, pivot):
        self._pivot = pivot

    @property
    def resolution(self) -> tuple:
        return self._resolution

    @resolution.setter
    def resolution(self, resolution):
        if self._resolution != resolution:
            self._resolution = resolution
            if self.window is not None:
                sdl2.SDL_FreeSurface(self.window)
            if self.renderer is not None:
                sdl2.SDL_DestroyRenderer(self.renderer)
            if self._cache["window_tx"] is not None:
                sdl2.SDL_DestroyTexture(self._cache["window_tx"])
            self.window = self.renderer = self._cache["window_tx"] = None

            self.window = sdl2.SDL_CreateRGBSurface(0, self._resolution[0], self._resolution[1], 32, 0, 0, 0, 0)
            self.renderer = sdl2.SDL_CreateSoftwareRenderer(self.window)

            self.object_render = False

    @property
    def logicalResolution(self) -> tuple:
        return self._resolution

    @logicalResolution.setter
    def logicalResolution(self, resolution):
        self._resolution = resolution



from ._sdl.sdl2 import SDL_Color
from .Log import LogError

__all__ = ["Color"]

class Color:
    def __init__(self, *color):
        self._r = self._g = self._b = self._a = 255
        self._SetColor(*color)

    def __repr__(self):
        return f"<%s.%s color=(%d, %d, %d, %d) at 0x%X>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self._r, self._g, self._b, self._a,
            id(self)
        )

    def _SetColor(self, *color):
        color = color[0] if len(color) == 1 else color
        if isinstance(color, (tuple, list)):
            if all(isinstance(c, int) for c in color):
                self.RGBA = color
            elif all(isinstance(c, float) for c in color):
                self.RGBA01 = color
            else:
                LogError(f"The value '{color}' is not recognized as a color")
        elif isinstance(color, int):
            self.Uint32 = color
        elif isinstance(color, str):
            self.Hex = color
        elif isinstance(color, SDL_Color):
            self.SDLColor = color
        else:
            LogError(f"The value '{color}' is not recognized as a color")

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, r):
        self._r = 0 if r < 0 else 255 if r > 255 else r

    @property
    def g(self):
        return self._g

    @g.setter
    def g(self, r):
        self._g = 0 if r < 0 else 255 if r > 255 else r

    @property
    def b(self):
        return self._b

    @b.setter
    def b(self, r):
        self._b = 0 if r < 0 else 255 if r > 255 else r

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, a):
        self._a = 0 if a < 0 else 255 if a > 255 else a

    @property
    def RGB(self):
        return self._r, self._g, self._b

    @RGB.setter
    def RGB(self, color):
        if isinstance(color, (tuple, list)) and all(isinstance(c, int) for c in color):
            self.r, self.g, self.b = color[0], color[1], color[2]
        else:
            LogError(f"The value '{color}' is not recognized as a RGB color")

    @property
    def RGBA(self):
        return self._r, self._g, self._b, self._a

    @RGBA.setter
    def RGBA(self, color):
        if isinstance(color, (tuple, list)) and all(isinstance(c, int) for c in color):
            self.RGB = color[0], color[1], color[2]
            self.a = self._a if len(color) < 4 else color[3]
        else:
            LogError(f"The value '{color}' is not recognized as a RGBA color")

    @property
    def RGB01(self):
        return round(self._r / 255, 4), round(self._g / 255, 4), round(self._b / 255, 4)

    @RGB01.setter
    def RGB01(self, color):
        if isinstance(color, (tuple, list)) and all(isinstance(c, float) for c in color):
            self.RGB = round(color[0] * 255), round(color[1] * 255), round(color[2] * 255)
        else:
            LogError(f"The value '{color}' is not recognized as a RGB01 color")

    @property
    def RGBA01(self):
        return round(self._r / 255, 4), round(self._g / 255, 4), round(self._b / 255, 4), round(self._a / 255, 4)

    @RGBA01.setter
    def RGBA01(self, color):
        if isinstance(color, (tuple, list)) and all(isinstance(c, float) for c in color):
            self.RGBA = round(color[0] * 255), round(color[1] * 255), round(color[2] * 255), round(color[3] * 255)
        else:
            LogError(f"The value '{color}' is not recognized as a RGBA01 color")

    @property
    def Uint32(self):
        return (self._r << 0) | (self._g << 8) | (self._b << 16) | (self._a << 24)

    @Uint32.setter
    def Uint32(self, color):
        if isinstance(color, int):
            self.RGBA = (color >> 0) & 0xFF, (color >> 8) & 0xFF, (color >> 16) & 0xFF, (color >> 24) & 0xFF
        else:
            LogError(f"The value '{color}' is not recognized as a color")

    @property
    def Hex(self):
        _color = f"{self._r:02X}{self._g:02X}{self._b:02X}{self._a:02X}"
        if all(_color[i] == _color[i + 1] for i in range(0, 8, 2)):
            _color = _color[::2]
        if len(_color) == 4 and _color[3] == 'F':
            _color = _color[:3]
        elif len(_color) == 8 and _color[6:] == 'FF':
            _color = _color[:6]
        return f"#{_color}"

    @Hex.setter
    def Hex(self, color):
        if isinstance(color, str):
            _color = color.lstrip("#")
            _color = "".join([char * 2 for char in _color]) if len(_color) == 3 or len(_color) == 4 else _color
            _color = _color + "FF" if len(_color) == 6 else _color
            try:
                self.RGBA = int(_color[0:2], 16), int(_color[2:4], 16), int(_color[4:6], 16), int(_color[6:8], 16)
            except ValueError:
                LogError(f"The value '{color}' is not recognized as a hexadecimal color")
        else:
            LogError(f"The value '{color}' is not recognized as a hexadecimal color")

    @property
    def Color(self):
        return self.RGBA

    @Color.setter
    def Color(self, *color):
        self._SetColor(*color)

    @property
    def SDLColor(self) -> SDL_Color:
        return SDL_Color(*self.RGBA)

    @SDLColor.setter
    def SDLColor(self, color: SDL_Color):
        if isinstance(color, SDL_Color):
            self.RGBA = color.r, color.g, color.b, color.a
        else:
            LogError(f"The value '{color}' is not recognized as a color SDL_Color")



from platform import python_version, version
from ._sdl.sdl2 import SDL_GetPlatform, SDL_GetVersion, SDL_GetNumRenderDrivers, SDL_RendererInfo, SDL_GetRenderDriverInfo
from ._sdl.sdlimage import IMG_Linked_Version
from ._sdl.sdlmixer import Mix_Linked_Version
from ._sdl.sdlttf import TTF_Linked_Version
from .Constants import RenderDriver as _RenderDriver
from .Log import LogCritical

__all__ = ["Platform"]

def compare_versions(version1, version2):
    for ver1, ver2 in zip(version1, version2):
        if ver1 < ver2:
            return -1
        elif ver1 > ver2:
            return 1
    return 0

class Platform:
    system = ""
    system_version = 0
    python_version = python_version()
    drivers = []

    class SDL:
        SDL_version_list = [0, 0, 0]
        SDL_version = ""

        SDLGFX_version_list = [0, 0, 0]
        SDLGFX_version = ""

        SDLIMAGE_version_list = [0, 0, 0]
        SDLIMAGE_version = ""

        SDLMIXER_version_list = [0, 0, 0]
        SDLMIXER_version = ""

        SDLTTF_version_list = [0, 0, 0]
        SDLTTF_version = ""

_systems = ["Windows", "Mac OS X", "Linux", "iOS", "Android"]
_compatible_systems = ["Windows"]

_min_versions_SDL = {"sdl": [2, 26, 0], "sdlgfx": [0, 0, 0], "sdlimage": [2, 0, 0], "sdlmixer": [2, 6, 0], "sdlttf": [2, 20, 0]}

Platform.system = SDL_GetPlatform()
if Platform.system not in _compatible_systems:
    LogCritical(f"incompatible {Platform.system} system")
if Platform.system.lower() == "windows":
    Platform.system_version = int(version().split(".")[2])

class RenderDriver:
    def __init__(self):
        self.id = self.name = self.hardware = self.software = self.bufferTexture = self.targetTexture = None

    def __repr__(self):
        return f"<%s.%s %s at 0x%X>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.name,
            id(self)
        )

for driver_index in range(SDL_GetNumRenderDrivers()):
    driver_info = SDL_RendererInfo()
    if SDL_GetRenderDriverInfo(driver_index, driver_info) == 0:
        _driver = RenderDriver()
        _driver.id = driver_index
        _driver.name = driver_info.name.decode('utf-8')
        _driver.hardware = bool(driver_info.flags & 0x00000002)
        _driver.software = bool(driver_info.flags & 0x00000001)
        _driver.bufferTexture = bool(driver_info.flags & 0x00000004)
        _driver.targetTexture = bool(driver_info.flags & 0x00000008)
        Platform.drivers.append(_driver)
        if _driver.name == "direct3d":
            _RenderDriver.Direct3d = _driver.id
        elif _driver.name == "direct3d11":
            _RenderDriver.Direct3d11 = _driver.id
        elif _driver.name == "direct3d12":
            _RenderDriver.Direct3d12 = _driver.id
        elif _driver.name == "opengl":
            _RenderDriver.OpenGL = _driver.id
        elif _driver.name == "software":
            _RenderDriver.Software = _driver.id

_SDL_Version = SDL_GetVersion()
Platform.SDL.SDL_version_list = [_SDL_Version.major, _SDL_Version.minor, _SDL_Version.patch]
Platform.SDL.SDL_version = ".".join(map(str, Platform.SDL.SDL_version_list))
if compare_versions(Platform.SDL.SDL_version_list, _min_versions_SDL["sdl"]) < 0:
    LogCritical()

Platform.SDL.SDLGFX_version_list = [0, 0, 0]
Platform.SDL.SDLGFX_version = ".".join(map(str, Platform.SDL.SDLGFX_version_list))
if compare_versions(Platform.SDL.SDLGFX_version_list, _min_versions_SDL["sdlgfx"]) < 0:
    LogCritical()

_SDL_Version = IMG_Linked_Version().contents
Platform.SDL.SDLIMAGE_version_list = [_SDL_Version.major, _SDL_Version.minor, _SDL_Version.patch]
Platform.SDL.SDLIMAGE_version = ".".join(map(str, Platform.SDL.SDLIMAGE_version_list))
if compare_versions(Platform.SDL.SDLIMAGE_version_list, _min_versions_SDL["sdlimage"]) < 0:
    LogCritical()

_SDL_Version = Mix_Linked_Version()
Platform.SDL.SDLMIXER_version_list = [_SDL_Version.major, _SDL_Version.minor, _SDL_Version.patch]
Platform.SDL.SDLMIXER_version = ".".join(map(str, Platform.SDL.SDLMIXER_version_list))
if compare_versions(Platform.SDL.SDLMIXER_version_list, _min_versions_SDL["sdlmixer"]) < 0:
    LogCritical()

_SDL_Version = TTF_Linked_Version()
Platform.SDL.SDLTTF_version_list = [_SDL_Version.major, _SDL_Version.minor, _SDL_Version.patch]
Platform.SDL.SDLTTF_version = ".".join(map(str, Platform.SDL.SDLTTF_version_list))
if compare_versions(Platform.SDL.SDLTTF_version_list, _min_versions_SDL["sdlttf"]) < 0:
    LogCritical()
